/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "acqiris.h"
#include "acqiris.moc"
#include <QtCore/QTimer>
#include <QtCore/QDateTime>

#include <QtCore/QThread>

#include <tiaction.h>

#ifdef HAVE_ACQIRIS
#include "acqiris_thread.h"
#include "acqiris_thread.moc"
#include "AcqirisD1Import.h"
#endif

#include <cmath>

Q_DECLARE_METATYPE( NSCore::AcqisitionDataSet )

namespace NSAcqiris {

	Acqiris::Acqiris( QObject* p ) : NSCore::QTreeItem( "acqiris", p ) {
		qDebug( "Acqiris::Acqiris()" );
#ifdef HAVE_ACQIRIS
		qDebug( " Acqiris-libs and -headers seemed to be found. This class will be more than a dummy." );
		ViInt32 nbrInstruments=0;
		ViString options = ( char* )"CAL=0";
		errorMessage( 0, AcqrsD1_multiInstrAutoDefine( options, &nbrInstruments ), "AcqrsD1_multiInstrAutoDefine" );
		qDebug( "found instruments: %li - creating the digitizers", (long)nbrInstruments );
		//options = (char*)"DMA=0";
		for ( long i=0; i<nbrInstruments; i++ ) {
			ViSession instrumentID=0;
			char name[ 20 ];
			sprintf( name, "PCI::INSTR%li", i );
			errorMessage( 0, AcqrsD1_InitWithOptions( name, VI_FALSE, VI_FALSE, options, &instrumentID ), "AcqrsD1_InitWithOptions" );
			_digitizers.push_back( new Digitizer( instrumentID, this, QString( "digitizer %1" ).arg( i ) ) );
		}
#endif
		qDebug( "initialized digitizers: %i", int( _digitizers.size() ) );
		//debugPrint();
	}

	Acqiris::~Acqiris() {
		qDebug( "Acqiris::~Acqiris()" );
		while ( _digitizers.size()>0 ) {
			delete *( _digitizers.begin() );
			_digitizers.pop_front();
		}
#ifdef HAVE_ACQIRIS
			errorMessage( 0, AcqrsD1_closeAll(), "AcqrsD1_closeAll" );
#endif
		qDebug( "Acqiris::~Acqiris() done." );
	}

	Digitizer* Acqiris::digitizer( int n ) {
		std::list<Digitizer*>::iterator it = _digitizers.begin();
		for ( int i=0; i<int( _digitizers.size() ) && it!=_digitizers.end(); ++i ) {
			if ( n == i )
				return ( *it );
			++it;
		}
		return 0;
	}


	Digitizer::Digitizer( quint32 instID, Acqiris* p, QString name )
		: NSCore::DataSetProducer( name, p )
		, _instrumentID( instID )
		, _c( new NSCore::TiParameter( "config", QVariant::Invalid, this ) )
#ifdef HAVE_ACQIRIS
		, _acqthread( new DigitizerThread( instID, _c, this ) )
#else
		, _acqthread( 0 )
#endif
		, _timer( new QTimer( this ) )
	{
#ifdef HAVE_ACQIRIS
		qRegisterMetaType<NSCore::AcqisitionDataSet>();

		_acqthread->start();

		for ( int i=0; i<QThread::idealThreadCount(); ++i ) {
			ConverterThread* tmp = new ConverterThread( _acqthread, this );
			tmp->start();
			connect( tmp, SIGNAL( newData( NSCore::AcqisitionDataSet ) ), this, SLOT( forwardData( NSCore::AcqisitionDataSet ) ) );
			_converters.append( tmp );
		}

		connect( _acqthread, SIGNAL( finished() ), this, SIGNAL( finished() ) );
#endif

		qDebug( "\tDigitizer::Digitizer( ViSession %i )", int( instID ) );

		_c->addParameter( "timeout", 2000 );
		_c[ "timeout" ]->setRange( 1, 3600000 );
		_c->addParameter( "forcetrigger", QVariant( false ) );

		new NSCore::TiAction( "Start", this, ":/icons/kde/cr22-action-player_start.png", this, SLOT( start() ) );
		new NSCore::TiAction( "Stop", this, ":/icons/kde/cr22-action-player_stop.png", this, SLOT( stop() ) );
		NSCore::TiParameter* singleShot = new NSCore::TiParameter( "Single Shot", false, this );
		connect( singleShot, SIGNAL( guiValueChanged() ), this, SLOT( updateSingleShot() ) );
		new NSCore::TiAction( "Calibrate", this, this, SLOT( calibrate() ) );
	}

	Digitizer::~Digitizer() {
		qDebug( "\tDigitizer::~Digitizer()" );
#ifdef HAVE_ACQIRIS
		_acqthread->stopThread();
		_acqthread->wait();
		for ( int i=0; i<_converters.size(); ++i ) {
			_converters.at( i )->stopThread();
			_converters.at( i )->wait();
		}
#endif
	}

	QString Digitizer::debug() const {
		return QString( "digitizer %1" ).arg( int( _instrumentID ) );
	}

	void Digitizer::updateSingleShot() {
#ifdef HAVE_ACQIRIS
		_acqthread->setSingleShot( getNamedChild<NSCore::TiParameter>( "Single Shot" )->toBool() );
#endif
	}

	int Digitizer::channels() {
#ifdef HAVE_ACQIRIS
		return _acqthread->channels();
#else
		return 0;
#endif
	}

	int Digitizer::externaltriggers() {
#ifdef HAVE_ACQIRIS
		return _acqthread->externalTriggers();
#else
		return 0;
#endif
	}

	void Digitizer::start() {
		qDebug( "\tDigitizer::start()" );
#ifdef HAVE_ACQIRIS
		_acqthread->startAcqisition();
#endif
	}

	void Digitizer::stop() {
		qDebug() << "Digitizer::stop()";
#ifdef HAVE_ACQIRIS
		_acqthread->stopAcqisition();
#endif
	}
	void Digitizer::abort() {
		qDebug() << "Digitizer::abort()";
#ifdef HAVE_ACQIRIS
		_acqthread->emptyQueue();
#endif
	}

	void Digitizer::calibrate() {
#ifdef HAVE_ACQIRIS
		_acqthread->scheduleCalibration();
#endif
	}

	void Digitizer::reset() {
#ifdef HAVE_ACQIRIS
		_acqthread->reset();
#endif
	}

	bool Digitizer::isRunning() {
#ifdef HAVE_ACQIRIS
		return ! _acqthread->queueIsEmpty();
#else
		return false;
#endif
	}

	void Digitizer::forwardData( NSCore::AcqisitionDataSet data ) {
		enqueue( data );
	}

#if 0
	void Digitizer::setSamplingInterval( const QString& n ) {
		qDebug() << "Digitizer::setSamplingInterval( const QString& " << n << " )";
		bool ok = false;
		double tmp = n.toDouble( &ok );
		if ( ok && tmp<1E-2 && tmp>=0.25E-9 ) {
			qDebug( "Setting samplingintervall to %e", tmp );
			_c[ "horizontal/samplinginterval" ]->setValue( tmp );
		}
	}
#endif

	QStringList Digitizer::triggersources() {
		QStringList ret;
#ifdef HAVE_ACQIRIS
		for ( int i=1; i<=_acqthread->channels(); i++ )
			ret<<"Channel "+QString::number( i );
		for ( int i=1; i<= _acqthread->externalTriggers(); i++ )
			ret<<"External "+QString::number( i );
		ret.sort();
#endif
		return ret;
	}

#if 0
	void Digitizer::triggersource( const QString &n ) {
		//qDebug( "\tDigitizer::triggersource( %s )", n.latin1() );
		if ( triggersources().contains( n ) ) {
			int channel=0;
			if ( n.contains( "Channel" )>0 )
				channel = n.right( -n.indexOf( " ", -1 ) ).toInt(); // XXX lastIndexOf?
			else if ( n.contains( "Extern" )>0 )
				channel = - n.right( -n.indexOf( " ", -1 ) ).toInt();
//			qDebug( "\nTrigger source will be %i\n", channel );
			_c[ "trigger/channel" ]->setValue( channel );
		}
	}
#endif

#if 0
	void Digitizer::triggerslope( const QString &n ) {
		if ( triggerslopes().contains( n ) ) {
			int set = 0;
			if ( n=="Positive" ) set = 0;
			if ( n=="Negative" ) set = 1;
			if ( n=="Out of Window" ) set = 2;
			if ( n=="Into Window" ) set = 3;
			if ( n=="HF Divide" ) set = 4;
			_c[ "trigger/slope" ]->setValue( set );
		}
	}
#endif

#if 0
	void Digitizer::triggercoupling( const QString &n ) {
		if ( triggercouplings().contains( n ) ) {
			int set = 0;
			if ( n=="DC" ) set = 0;
			if ( n=="AC" ) set = 1;
			if ( n=="HF Reject" ) set = 2;
			if ( n=="DC 50Ohm" ) set = 3;
			if ( n=="AC 50Ohm" ) set = 4;
			_c[ "trigger/coupling" ]->setValue( set );
		}
	}
#endif

#if 0
	double Digitizer::fullscale( QString n ) const {
		float v=5.0;
		bool ok=false;
		if ( n.contains( "mV" ) ) {
			v = n.left( n.indexOf( "mV" ) ).toFloat( &ok )/1000;
		} else if ( n.contains( "V" ) ) {
			v = n.left( n.indexOf( "V" ) ).toFloat( &ok );
		}
		if ( ok ) {
			return v;
		}
		return 5.0;
	}
#endif


}; // namespace NSAcqiris
