/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ACQIRIS_THREAD_H
#define ACQIRIS_THREAD_H

#include <QtCore/QThread>
#include <QtCore/QMultiMap>
#include <QtCore/QSemaphore>
#include <QtCore/QReadWriteLock>
#include <QtCore/QTime>

#include <queue>

#include <datatypes.h>
#include <qsemaphore64.h>

#include <tiparameter.h>
#include <tienum.h>

#ifdef HAVE_ACQIRIS
#include "AcqirisD1Import.h"
#endif

#include <cmath>

#include <unistd.h>
#include <sys/syscall.h>
#include <sys/prctl.h>

namespace NSAcqiris {

	qint32 errorMessage( const quint32 _inst, const qint32 status, const QString& comment="" ) {
		if ( status != 0 ) {
			ViChar msg[ 500 ];
			AcqrsD1_errorMessage( _inst, status, msg );
			qDebug() << QString( " Last call %1returned: %2" ).arg( comment.isEmpty()?"":"( "+comment+" ) " ).arg( msg );
		}
		return status;
	}

	struct filters_t {
		bool min, max;
		double min_min, min_max, max_min, max_max;
	};
	struct DigitizerChannelData {
		ViAddr dataArray;
		AqDataDescriptor* dataDesc;
		ViAddr segDescArray;
		struct filters_t filter;
		double rawmin, rawmax;
	};
	class DigitizerData {
	public:
		/// Number of channels in the dataset
		int channels;
		/// Number of segments in the dataset
		int segments;
		/// Number of bytes to make on sample on one channel
		int dataBytes;
		/// Overall size of the datablock of one channel (including overhead)
		int fullDataSize;

		std::vector<DigitizerChannelData> data;

	private:
		friend class DigitizerDataPool;

		DigitizerData( int ch, int seg, int dS, int dB ) : channels( ch ), segments( seg ), dataBytes( dB ), fullDataSize( dS ) {
			for ( int i=0; i<channels; ++i ) {
				DigitizerChannelData tmp;
				tmp.dataArray = new char[ dS ];
				tmp.segDescArray = new AqSegmentDescriptor[ segments ];
				tmp.dataDesc = new AqDataDescriptor;
				tmp.filter.min = tmp.filter.max = false;
				data.push_back( tmp );
			}
		}
		~DigitizerData() {
			while ( ! data.empty() ) {
				delete[] ( char* )data.back().dataArray;
				delete[] static_cast<AqSegmentDescriptor*>( data.back().segDescArray );
				delete data.back().dataDesc;
				data.pop_back();
			}
		}
	};
	class DigitizerDataPool {
	private:
		DigitizerDataPool() : _poolLock( new QMutex() ) {
			qDebug() << "DigitizerDataPool::DigitizerDataPool()";
		}
		~DigitizerDataPool() {
		}

		/// Pool
		int _lastUsedSize;
		QList<DigitizerData*> _pool;
		QMutex* _poolLock;
		QSemaphore64 _poolSemaphore;

	public:
		static DigitizerDataPool* the() {
			static DigitizerDataPool* me = 0;
			if ( ! me ) {
				me = new DigitizerDataPool();
				QFile* f = new QFile( "/proc/meminfo" );
				if ( f->open( QIODevice::ReadOnly ) ) {
					long long int globmem;
					qDebug() << "Successfully opened /proc/meminfo";
					QString tmp = f->readAll();
					globmem = tmp.section( "\r", 0 ).section( " ", 1, 1, QString::SectionSkipEmpty ).toLongLong() * 1024 / 4;
					qDebug() << " Global memory is" << globmem << "bytes, will reserve about a quarter of it...";
					f->close();
					me->_poolSemaphore.release( globmem );
					qDebug( " Reserved %lli bytes for the buffer. That is %lli kB, %lli MB", globmem, globmem/1024, globmem/1024/1024 );
					qDebug() << " available:" << me->_poolSemaphore.available();
				} else {
					me->_poolSemaphore.release( int(pow(2, 16)) );
				}
				delete f;
			}
			return me;
		}

		DigitizerData* getData( int ch, int seg, int dS, int dB ) {
			//qDebug() << "DigitizerDataPool::getData(" << ch << "," << seg << "," << dS << "," << dB << ")";
			//qDebug() << " available:" << _poolSemaphore.available();
			_poolSemaphore.acquire( ch * dS );
			int mysize = ch * dS;
			if ( _lastUsedSize != mysize ) {
				_lastUsedSize = mysize;
				while ( _pool.size() > 0 )
					delete _pool.takeFirst();
			}
			if ( _pool.size() > 0 ) {
				// Q: needs the lock?
				QMutexLocker lock( _poolLock );
				DigitizerData* ret = _pool.takeFirst();
				ret->channels = ch;
				ret->segments = seg;
				return ret;
			}
			return new DigitizerData( ch, seg, dS, dB );
		}
		void deleteData( DigitizerData* data ) {
			//qDebug() << "DigitizerDataPool::deleteData(" << data << ")";
			_poolSemaphore.release( data->channels * data->fullDataSize );
			if ( data->channels * data->fullDataSize == _lastUsedSize && _pool.size() < 1000 ) {
				QMutexLocker lock( _poolLock );
				_pool.append( data );
			} else {
				delete data;
			}
		}
	};

	class DigitizerThread : public QThread
	{
		Q_OBJECT

		quint32 _inst;

		bool _run, _quit, _single, _verticalChanged, _horizontalChanged, _triggerChanged, _scheduleCalibration, _filterChanged, _externalChanged, _convertersChanged;

		qint32 _totalChannels, _activeChannels, _activeChannelCount, _externaltriggers;

		NSCore::TiParameter* _config;

		std::queue<DigitizerData*> _data;
		QMutex* _datamutex;

		time_t _temperatur_time;
		NSCore::TiParameter* _temperatur_param;

		qint32 _datatype_to_read;

		time_t _statistics_time;
		NSCore::TiParameter* _statistics_cps_param;
		double _statistics_cps;

		NSCore::TiParameter* _acquisitionscounter;
		NSCore::TiParameter* _acquisitionscounter_max;

		QString _devicename;
		Digitizer* _digitizer;

		QStringList _fullscales, _couplings;

		public:
			DigitizerThread( ViSession s, NSCore::TiParameter* config, Digitizer* p )
				: QThread( p )
				, _inst( s )
				, _run( false )
				, _quit( false )
				, _single( false )
				, _verticalChanged( true )
				, _horizontalChanged( true )
				, _triggerChanged( true )
				, _scheduleCalibration( true )
				, _filterChanged( true )
				, _externalChanged( true )
				, _convertersChanged( true )
				, _totalChannels( 0 )
				, _activeChannels( 0 )
				, _config( config )
				, _datamutex( new QMutex )
				, _datatype_to_read( 0 )
				, _statistics_cps( 0.0 )
				, _digitizer( p )
			{
				qDebug() << "DigitizerThread::DigitizerThread(" << s << "," << p << ")";

				char* options = new char[ 64 ];
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"Options", options ), "AcqrsD1_getInstrumentInfo" );
				qDebug() << " The digitizer seems to have the following options installed:" << options;
				qint32 tmp;
				errorMessage( AcqrsD1_getInstrumentData( _inst, options, ( ViInt32* )&tmp, ( ViInt32* )&tmp, ( ViInt32* )&tmp ), "AcqrsD1_getInstrumentData" );
				_devicename = options;
				qDebug() << " Device name is" << _devicename;
				delete[] options;
				options = 0;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"MaxSamplesPerChannel", &tmp ) );
				qDebug() << " Maximum number of samples per channel is:" << tmp;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrADCBits", &tmp ) );
				qDebug() << " Number of ADC-bits:" << tmp;
				if ( tmp > 8 )
					_datatype_to_read = 1;

				qDebug() << endl;

				_config->addParameter( "Stop after N acquisitions", QVariant( 0 ) );
				_acquisitionscounter_max = _config->getParameter( "Stop after N acquisitions" );
				_acquisitionscounter_max->setRange( 0, int( 10E12 ) );
				_config->addParameter( "Acquisitions Counter", QVariant::ULongLong );
				_acquisitionscounter = _config->getParameter( "Acquisitions Counter" );
				_acquisitionscounter->setGuiHint( NSCore::TiParameter::Counter );

				int converters, usedChannels;
				errorMessage( AcqrsD1_getChannelCombination( _inst, (ViInt32*)&converters, (ViInt32*)&usedChannels ), "AcqrsD1_getChannelCombination" );
				_config->addParameter( "Converters per Channel", converters );
				_config->getParameter( "Converters per Channel" )->setRange( 1, 4 );
				connect( _config->getParameter( "Converters per Channel" ), SIGNAL( guiValueChanged() ), this, SLOT( convertersChanged() ) );

				_fullscales << "0mV" << "50mV" << "100mV" << "200mV" << "500mV" << "1V" << "2V" << "5V";
				_couplings << "Ground" << "DC 1MOhm" << "AC 1MOhm" << "DC 50Ohm" << "AC 50Ohm";
				if ( _devicename == "DC440" ) {
					_couplings[3] = "Standard Input";
					_couplings[4] = "HF Input";
					_fullscales.clear();
					_fullscales << "250mV" << "500mV" << "1V" << "2V" << "5V" << "10V";
				}
				int channels;
				errorMessage( AcqrsD1_getNbrChannels( _inst, ( ViInt32* )&channels ), "AcqrsD1_getNbrChannels" );
				initChannels( channels );
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrExternalTriggers", &_externaltriggers ), "getExternalTriggers" );
				for ( int i=-1; i>=-_externaltriggers; i-=2 ) {
					_config->addParameter( QString( "Channel %1" ).arg( i ), QVariant::Invalid );
					NSCore::TiParameter* tmp = _config->getParameter( QString( "Channel %1" ).arg( i ) );
					new NSCore::TiEnum( "fullscale", _fullscales, "5V", tmp );
					new NSCore::TiEnum( "bandwidth", QStringList() << "No Bandlimit" << "25MHz" << "700MHz" << "200MHz" << "20MHz" << "35MHz", 0, tmp );
					connect( tmp, SIGNAL( guiValueChanged() ), this, SLOT( verticalChanged() ) );
				}

				_config->addParameter( "horizontal/samplinginterval", 1/4.0, "Sampling interval in units of nanoseconds" );
				if ( _devicename == "DC440" )
					_config->getParameter( "horizontal/samplinginterval" )->setRange( 1000.0/420, 1000 );
				_config->addParameter( "horizontal/samples", 100 );
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"MaxSamplesPerChannel", &tmp ) );
				_config->getParameter( "horizontal/samples" )->setRange( 16, tmp );
				_config->addParameter( "horizontal/segments", 1, "Number of segments to acquire in one run" );
				_config->getParameter( "horizontal/segments" )->setRange( 1, int( pow( 2, 16 ) ) );
				_config->addParameter( "horizontal/delaytime", 0.0, "The time the trigger is delayed (in nanoseconds)." );
				_config->getParameter( "horizontal/delaytime" )->setRange( -100000.0, 100000.0 );
				connect( _config->getParameter( "horizontal" ), SIGNAL( guiValueChanged() ), this, SLOT( horizontalChanged() ) );

				_config->addParameter( "trigger/channel", 1, "Positive numbers for the input-channels, negative numbers for the external triggers." ); // TODO enum
				_config->getParameter( "trigger/channel" )->setRange( int( -_externaltriggers ), int( _totalChannels ) );
				new NSCore::TiEnum( "slope", QStringList() << "Positive" << "Negative" << "Out of Window" << "Into Window" << "HF Divide", 0, _config->getParameter( "trigger" ) );
				_config->addParameter( "trigger/level", 0.0 );
				_config->getParameter( "trigger/level" )->setGuiHint( NSCore::TiParameter::ReadAllowed );
				_config->addParameter( "trigger/levelintern", 0.0, "The trigger-level for internal triggermode in percent." );
				_config->getParameter( "trigger/levelintern" )->setRange( -100.0, 100.0 );
				connect( _config->getParameter( "trigger/levelintern" ), SIGNAL( guiValueChanged() ), this, SLOT( triggerLevelInternChanged() ) );
				_config->addParameter( "trigger/levelextern", 0.0, "The trigger-level for external triggers in Volts." );
				_config->getParameter( "trigger/levelextern" )->setRange( -5.0, 5.0 );
				connect( _config->getParameter( "trigger/levelextern" ), SIGNAL( guiValueChanged() ), this, SLOT( triggerLevelExternChanged() ) );
				_config->addParameter( "trigger/level2", 0.0 );
				_config->getParameter( "trigger/level2" )->setRange( -100.0, 100.0 );
				new NSCore::TiEnum( "coupling", QStringList() << "DC 1MOhm" << "AC 1MOhm" << "HF Reject" << "DC 50Ohm" << "AC 50Ohm", 0, _config->getParameter( "trigger" ) );
				connect( _config->getParameter( "trigger" ), SIGNAL( guiValueChanged() ), this, SLOT( triggerChanged() ) );

				/// external io
				_config->addParameter( "external io", QVariant::Invalid );
				_config->addParameter( "external io/io a", QVariant::Int, "0: Disable <br>Input:<br>6: Enable Trigger <br>8: Prepare for trigger <br>Output:<br>19: 10MHz Reference Clock <br>20: Acquisition skips to next segment <br>21: Acquisition is active <br>22: Trigger is armed" );
				_config->addParameter( "external io/level a", QVariant::Int );
				_config->addParameter( "external io/io b", QVariant::Int, "0: Disable <br>Input:<br>6: Enable Trigger <br>8: Prepare for trigger <br>Output:<br>19: 10MHz Reference Clock <br>20: Acquisition skips to next segment <br>21: Acquisition is active <br>22: Trigger is armed" );
				_config->addParameter( "external io/level b", QVariant::Int );
				connect( _config->getParameter( "external io" ), SIGNAL( guiValueChanged() ), this, SLOT( externalChanged() ) );

				_config->addParameter( "status/temperatur", 0.0 );
				_temperatur_param = _config->getParameter( "status/temperatur" );
				_temperatur_param->setGuiHint( NSCore::TiParameter::ReadAllowed );
				_temperatur_time = time( 0 );
				_config->addParameter( "status/counts per second", 0.0 );
				_statistics_cps_param = _config->getParameter( "status/counts per second" );
				_statistics_cps_param->setGuiHint( NSCore::TiParameter::ReadAllowed );
				_statistics_time = time( 0 );

			}
			~DigitizerThread() {
				qDebug() << "DigitizerThread::~DigitizerThread()";
			}

			void stopThread() { _quit = true; }

			void startAcqisition() {
				qDebug() << "DigitizerThread::startAcqisition() setting _run=true";
				_run = true;
			}
			void stopAcqisition() {
				qDebug() << "DigitizerThread::stopAcqisition() setting _run=false";
				_run = false;
			}
			void setSingleShot( bool n ) { _single = n; }
			void scheduleCalibration() { _scheduleCalibration = true; }

			int channels() const { return _activeChannels; }
			int externalTriggers() const { return _externaltriggers; }

			void reset() {
				_config->getParameter("Acquisitions Counter")->reset();
			}

			/**
			 * @brief Get and release data from the queue
			 */
			DigitizerData* first() {
				QMutexLocker lock( _datamutex );
				DigitizerData* tmp = 0;
				if ( !_data.empty() ) {
					tmp = _data.front();
					_data.pop();
				}
				return tmp;
			 }

			/**
			 * @brief Empty the queue
			 */
			void emptyQueue() {
				QMutexLocker lock( _datamutex );
				while ( !_data.empty() ) {
					_data.pop();
				}
			}

			bool queueIsEmpty() {
				return _data.empty();
			}

		signals:
			void finished();

		private slots:
			/**
			 * @{
			 *
			 * @brief Settings have changed.
			 */
			void verticalChanged()   { _verticalChanged = true; }
			void horizontalChanged() { _horizontalChanged = true; }
			void triggerChanged()    { _triggerChanged = true; }
			void filterChanged()     { _filterChanged = true; }
			void externalChanged()   { _externalChanged = true; }
			void convertersChanged() { _convertersChanged = true; }
			// @}

			/// Set the trigger level from the internal trigger level.
			void triggerLevelInternChanged() {
				//qDebug() << "DigitizerThread::triggerLevelInternChanged()";
				if ( _config->getParameter( "trigger/channel" )->toInt() > 0 )
					_config->getParameter( "trigger/level" )->setValue( _config->getParameter( "trigger/levelintern" )->toDouble() );
			}
			/// Set the trigger level from the external trigger level.
			void triggerLevelExternChanged() {
				//qDebug() << "DigitizerThread::triggerLevelExternChanged()";
				if ( _config->getParameter( "trigger/channel" )->toInt() < 0 )
					_config->getParameter( "trigger/level" )->setValue( _config->getParameter( "trigger/levelextern" )->toDouble() * 1000 );
			}

			void initChannels( int channels ) {
				qDebug( "initChannels: old %i, new %i", _totalChannels, channels );
				while( _totalChannels > channels ) {
					delete _config->getParameter( QString( "Channel %1" ).arg( _totalChannels+1 ) );
					qDebug() << " ! Removed one channel.";
					_totalChannels--;
				}
				while( _totalChannels < channels ) {
					_config->addParameter( QString( "Channel %1" ).arg( _totalChannels+1 ), QVariant::Invalid );
					NSCore::TiParameter* tmp = _config->getParameter( QString( "Channel %1" ).arg( _totalChannels+1 ) );
					new NSCore::TiEnum( "fullscale", _fullscales, "5V", tmp );
					tmp->addParameter( "offset", 0.0 );
					tmp->getParameter( "offset" )->setRange( -5.0, 5.0 );
					new NSCore::TiEnum( "coupling", _couplings, 3, tmp );
					new NSCore::TiEnum( "bandwidth", QStringList() << "No Bandlimit" << "25MHz" << "700MHz" << "200MHz" << "20MHz" << "35MHz", 0, tmp );
					tmp->addParameter( "Minimum Filter", false );
					tmp->addParameter( "Minimum Filter/min", -5.0 );
					tmp->addParameter( "Minimum Filter/max", 5.0 );
					tmp->addParameter( "Maximum Filter", false );
					tmp->addParameter( "Maximum Filter/min", -5.0 );
					tmp->addParameter( "Maximum Filter/max", 5.0 );
					connect( tmp, SIGNAL( guiValueChanged() ), this, SLOT( filterChanged() ) );
					connect( tmp, SIGNAL( guiValueChanged() ), this, SLOT( verticalChanged() ) );
					qDebug() << " * Created one channel.";
					_totalChannels++;
				}
			}

		private:
			/**
			 * @brief The real working function.
			 *
			 * Is run in the new thread...
			 */
			void run() {
				prctl(PR_SET_NAME, "DigitizerThread");

				qDebug( " ! The DigitizerThread is now running with pid %lu!", syscall(SYS_gettid) );

				//errorMessage( AcqrsD1_calibrate( _inst ), "AcqrsD1_calibrate" );

				errorMessage( AcqrsD1_configControlIO( _inst, 1, 19, 0, 0.0 ), "AcqrsD1_configControlIO" );

				QVector<filters_t> filters;

				uint cT;

				long currentpad = 0;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"TbSegmentPad", &currentpad ), "AcqrsD1_getInstrumentInfo" );

				long hwDevices = 0;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrModulesInInstrument", &hwDevices ), "AcqrsD1_getInstrumentInfo" );

				DigitizerDataPool::the();

				while ( !_quit ) {

					cT = time( 0 );

					/*
					 * Read the temperature of the first physical device every 5 seconds.
					 */
					if ( cT - _temperatur_time > 5 ) {
						qint32 temp;
						errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"Temperature 0", &temp ) );
						_temperatur_param->setValue( temp*1.0 );
						_temperatur_time = cT;
					}
					/*
					 * Do statistics every second.
					 */
					if ( cT - _statistics_time > 1 ) {
						_statistics_cps_param->setValue( _statistics_cps );
						_statistics_cps = 0.0;
						_statistics_time = cT;
					}

					if ( _run == false ) {
						/*
						 * If there is no acquisition running, we can just sleep...
						 */
						usleep( 1000 );
					} else {
						/*
						 * There are a lot of things to do when an acquisition should be done.
						 */

						//qDebug() << " * DigitizerThread: checking requisits";
						qint32 status;

						// Check if the channel combination changed
						if ( _convertersChanged ) {
							int channelsPerDevice = _totalChannels / hwDevices;
							int converters = 1, usedChannels=0;
							qDebug( " TotalChannels %i / hwDevices %li = channelsPerDevice %i", _totalChannels, hwDevices, channelsPerDevice );
							converters = _config->getParameter( "Converters per Channel" )->toInt();
							for ( int i=0; i<channelsPerDevice/converters; i++ )
								usedChannels |= (1<<i);
							qDebug( " New converters %i with  usedChannels 0x%x", converters, usedChannels );
							errorMessage( AcqrsD1_configChannelCombination( _inst, converters, usedChannels ), "AcqrsD1_configChannelCombination" );
							//_activeChannels = _totalChannels/converters;
							_activeChannelCount = _totalChannels/converters;
							//_activeChannels = _totalChannels;
							_activeChannels = 0;
							for( int i=0; i<hwDevices-1; ++i ) {
								_activeChannels |= (usedChannels<<(i*4));
							}
							qDebug( " Active Channels %x", _activeChannels );
							for( int i=1; i<=_totalChannels; ++i ) {
								if( _activeChannels & (1<<(i-1)) )
									_config->getParameter( QString( "Channel %1" ).arg( i ) )->setGuiHint( NSCore::TiParameter::ReadWrite );
								else
									_config->getParameter( QString( "Channel %1" ).arg( i ) )->setGuiHint( NSCore::TiParameter::ReadAllowed );
							}
							_verticalChanged = true;
							_convertersChanged = false;
							_scheduleCalibration = true;
						}
						// Check wether the vertical parameters changed and set them accordingly.
						if ( _verticalChanged ) {
							qDebug() << " Configuring vertical properties";
							for ( int i=-_externaltriggers; i<=_totalChannels; ++i )
								if ( i!= 0 )
									configureChannel( i );
							_verticalChanged = false;
						}
						// Then check wether the horizontal parameters changed and set them accordingly.
						if ( _horizontalChanged ) {
							qDebug() << " Configuring horizontal properties";
							status = errorMessage( AcqrsD1_configHorizontal( _inst,
								_config->getParameter( "horizontal/samplinginterval" )->toDouble() * 1E-9,
								_config->getParameter( "horizontal/delaytime" )->toDouble() * 1E-9 ),
								"AcqrsD1_configHorizontal" );
							if ( status != 0 ) {
								double delay, interval;
								AcqrsD1_getHorizontal( _inst, &interval, &delay );
								qDebug( " New values are: interval=%g delay=%g", interval, delay );
							}
							status = errorMessage( AcqrsD1_configMemory( _inst,
								_config->getParameter( "horizontal/samples" )->toInt(),
								_config->getParameter( "horizontal/segments" )->toInt() ),
								"AcqrsD1_configMemory" );
							if ( status != 0 ) {
								ViInt32 samples, segments;
								AcqrsD1_getMemory( _inst, &samples, &segments );
								qDebug( " New values are: samples=%li segments=%li", (long)samples, (long)segments );
							}
							_horizontalChanged = false;
						}
						// The trigger settings can also change and need to be set.
						if ( _triggerChanged ) {
							qDebug() << " Configuring trigger properties";
							errorMessage( AcqrsD1_configTrigClass( _inst, 0, triggerSourceToPattern( _config->getParameter( "trigger/channel" )->toInt() ), 0, 0, 0, 0 ) );
							errorMessage( AcqrsD1_configTrigSource( _inst,
								_config->getParameter( "trigger/channel" )->toInt(),
								_config->getNamedChild<NSCore::TiEnum>( "trigger/coupling" )->toInt(),
								_config->getNamedChild<NSCore::TiEnum>( "trigger/slope" )->toInt(),
								_config->getParameter( "trigger/level" )->toDouble(),
								_config->getParameter( "trigger/level2" )->toDouble() ) );
							_triggerChanged = false;
						}
						// Did a filter-setting change?
						if ( _filterChanged ) {
							qDebug() << " Configuring filter properties";
							if ( filters.size() != _totalChannels ) filters.resize( _totalChannels );
							for ( int i=0; i<_totalChannels; ++i ) {
								filters[ i ].min = _config->getParameter( QString( "Channel %1/Minimum Filter" ).arg( i+1 ) )->toBool();
								filters[ i ].min_min = _config->getParameter( QString( "Channel %1/Minimum Filter/min" ).arg( i+1 ) )->toDouble();
								filters[ i ].min_max = _config->getParameter( QString( "Channel %1/Minimum Filter/max" ).arg( i+1 ) )->toDouble();
								filters[ i ].max = _config->getParameter( QString( "Channel %1/Maximum Filter" ).arg( i+1 ) )->toBool();
								filters[ i ].max_min = _config->getParameter( QString( "Channel %1/Maximum Filter/min" ).arg( i+1 ) )->toDouble();
								filters[ i ].max_max = _config->getParameter( QString( "Channel %1/Maximum Filter/max" ).arg( i+1 ) )->toDouble();
							}
							_filterChanged = false;
						}
						// Did the external io config change?
						if ( _externalChanged ) {
							qDebug() << " Configuring external properties";
							errorMessage( AcqrsD1_configControlIO( _inst, 1, _config->getParameter( "external io/io a" )->toInt(), _config->getParameter( "external io/level a" )->toInt(), 0), "AcqrsD1_configControlIO" );
							errorMessage( AcqrsD1_configControlIO( _inst, 2, _config->getParameter( "external io/io b" )->toInt(), _config->getParameter( "external io/level b" )->toInt(), 0), "AcqrsD1_configControlIO" );
							_externalChanged = false;
						}

						// Is a calibration wanted?
						if ( _scheduleCalibration ) {
							qDebug() << " Running calibration";
							errorMessage( AcqrsD1_calibrate( _inst ), "AcqrsD1_calibrate" );
							_scheduleCalibration = false;
							qDebug() << " Calibration finished";
						}

						//qDebug() << " * DigitizerThread: starting the acquisition";

						/*
						 * Do the real acquisition.
						 */
						errorMessage( AcqrsD1_acquire( _inst ), "AcqrsD1_acquire" );
						{
						long oldpad = currentpad;
						errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"TbSegmentPad", &currentpad ), "AcqrsD1_getInstrumentInfo" );
						if ( oldpad != currentpad ) qDebug( " Acquisition started: SegmentPad changed to %li.", currentpad );
						}
						status = errorMessage( AcqrsD1_waitForEndOfAcquisition( _inst, _config->getParameter( "timeout" )->toInt() ), "AcqrsD1_waitForEndOfAcquisition" );

						//qDebug() << " * DigitizerThread: waitForEndOfAcquisition returned" << status;

						// Timeout-handling
						if ( status == signed( ACQIRIS_ERROR_ACQ_TIMEOUT ) ) {
							status = AcqrsD1_stopAcquisition( _inst );
							errorMessage( status );
							/*if ( _config->getParameter( "forcetrigger" )->toBool() )
								errorMessage( AcqrsD1_waitForEndOfAcquisition( _inst, 500 ) );
							else
								_run = false;*/
						}

						ViBoolean tmp;
						errorMessage( AcqrsD1_acqDone( _inst, &tmp ), "AcqrsD1_acqDone" );
						//qDebug() << " * DigitizerThread: acqDone returns" << bool( tmp );

						/*
						 * Get the data from the digitizer and store it in the outgoing buffer.
						 */

						AqReadParameters readPar;
						readPar.flags = 0;
						readPar.dataType = _datatype_to_read;
						readPar.readMode = 1;
						readPar.firstSegment = 0;
						readPar.firstSampleInSeg = 0;
						errorMessage( AcqrsD1_getMemory( _inst, &( readPar.nbrSamplesInSeg ), &( readPar.nbrSegments ) ), "AcqrsD1_getMemory" );
						errorMessage( AcqrsD1_reportNbrAcquiredSegments( _inst, &( readPar.nbrSegments ) ), "AcqrsD1_reportNbrAcquiredSegments" );
						readPar.nbrSegments = qMax( readPar.nbrSegments, ViInt32(0L) );
						//qDebug( "Returned number of finished segments: %li", readPar.nbrSegments );
						readPar.segmentOffset = readPar.nbrSamplesInSeg;
						readPar.segDescArraySize = sizeof( AqSegmentDescriptor ) * readPar.nbrSegments;
						readPar.dataArraySize = ( readPar.nbrSamplesInSeg * int( pow( 2, readPar.dataType ) ) + currentpad ) * ( 1+readPar.nbrSegments );
						readPar.dataArraySize = qMax( readPar.dataArraySize, ViInt32(( readPar.nbrSamplesInSeg + currentpad ) * ( readPar.nbrSegments + 1 ) * int( pow( 2, readPar.dataType ) )) );
						//qDebug() << " Will allocate buffer of size" << readPar.dataArraySize << "bytes (=" << readPar.nbrSamplesInSeg << "*" << int( pow( 2, readPar.dataType ) ) << "+" << currentpad << "*" << ( 1+readPar.nbrSegments ) << ")";

						DigitizerData* data = DigitizerDataPool::the()->getData( _activeChannelCount, qMax(ViInt32(0L), readPar.nbrSegments), readPar.dataArraySize, pow( 2, readPar.dataType ) );

						int c = 0;
						for ( int i=0; i<_totalChannels; ++i ) {
							//qDebug( "Checking channel %i for data, that is mask %x in %x", i, (1<<(i%(hwDevices-1))), _activeChannels );
							if( _activeChannels&(1<<(i%(hwDevices-1))) ) {
								status = errorMessage( AcqrsD1_readData( _inst, i+1, &readPar, data->data[ c ].dataArray, data->data[ c ].dataDesc, data->data[ c ].segDescArray ), QString( "AcqrsD1_readData (ch=%1)" ).arg( i+1 ) );
								data->data[ c ].filter = filters[ i ];
								c++;
							}
						}
						if ( status != 0 ) {
							_run = false;
							emit finished();
						}

						if ( !_digitizer->consumersIsEmpty() ) {
							QMutexLocker lock( _datamutex );
							//_data.enqueue( data );
							_data.push( data );
						} else
							DigitizerDataPool::the()->deleteData( data );

						_statistics_cps += readPar.nbrSegments;

						/*
						 * Check wether a next acquisition is wanted or not.
						 */
						_acquisitionscounter->setValue( QVariant( int( _acquisitionscounter->toInt() + readPar.nbrSegments ) ) );
						if ( _acquisitionscounter_max->toInt() > 0
								&& _acquisitionscounter->toInt() > _acquisitionscounter_max->toInt() ) {
							_run = false;
							emit finished();
						}

						if ( _single || _quit )
							_run = false;
					}
				}
				qDebug() << "DigitizerThread finished!" << endl;
			}

			/**
			 * @brief Configuration of channels.
			 *
			 * Fits for both normal channels and external trigger inputs.
			 */
			void configureChannel( int ch ) {
				if ( _config->getNamedChild( QString( "Channel %1" ).arg( ch ) ) == 0 )
					return;
				QString path = QString( "Channel %1/" ).arg( ch );
				ViReal64 fs_in = _config->getNamedChild( path + "fullscale" )?fullscale( _config->getNamedChild<NSCore::TiEnum>( path + "fullscale" )->toString() ):0.0;
				ViReal64 offset_in = _config->getNamedChild( path + "offset" )?_config->getNamedChild<NSCore::TiParameter>( path + "offset" )->toDouble():0.;
				ViInt32 coupling_in = _config->getNamedChild<NSCore::TiEnum>( path + "coupling" )?_config->getNamedChild<NSCore::TiEnum>( path + "coupling" )->toInt():3;
				ViInt32 bandwidth_in = _config->getNamedChild<NSCore::TiEnum>( path + "bandwidth" )?_config->getNamedChild<NSCore::TiEnum>( path + "bandwidth" )->toInt():0;
				errorMessage( AcqrsD1_configVertical( _inst, ch, fs_in, offset_in, coupling_in, bandwidth_in ), QString( "configVertical( %1 )" ).arg( ch ) );

				ViReal64 fs_out = 0, offset_out = 0;
				ViInt32 coupling_out = 3, bandwidth_out = 0;
				errorMessage( AcqrsD1_getVertical( _inst, ch, &fs_out, &offset_out, &coupling_out, &bandwidth_out ), QString( "getVertical( %1 )" ).arg( ch ) );
				if( fs_out != fs_in ) {
					qDebug() << " Fullscale is now" << fs_out << "instead of" << fs_in;
					QString fs = QString::number( int( fs_out ) );
					if ( fs_out < 1 )
						fs = QString::number( int( fs_out*1000 ) ) + "m";
					fs += "V";
					qDebug() << "  which becomes '" << fs << "'";
					_config->getNamedChild<NSCore::TiEnum>( path + "fullscale" )->setValue( fs );
				}
				if( offset_out != offset_in ) {
					qDebug() << " Offset is now" << offset_out << "instead of" << offset_in;
					_config->getNamedChild<NSCore::TiParameter>( path + "offset" )->setValue( double( offset_out ) );
				}
				if( coupling_out != coupling_in ) {
					qDebug() << " Coupling is now" << coupling_out << "instead of" << coupling_in;
					_config->getNamedChild<NSCore::TiEnum>( path + "coupling" )->setValue( int(coupling_out) );
				}
				if( bandwidth_out != bandwidth_in ) {
					qDebug() << " Bandwidth is now" << bandwidth_out << "instead of" << bandwidth_in;
					_config->getNamedChild<NSCore::TiEnum>( path + "bandwidth" )->setValue( int(bandwidth_out) );
				}
			}

			/**
			 * @brief Convert a fullscale-string to its double value.
			 */
			double fullscale( QString n ) const {
				float v=5.0;
				bool ok=false;
				if ( n.contains( "mV" ) )
					v = n.left( n.indexOf( "mV" ) ).toFloat( &ok )/1000;
				else if ( n.contains( "V" ) )
					v = n.left( n.indexOf( "V" ) ).toFloat( &ok );
				if ( ok )
					return v;
				return 5.0;
			}
			/**
			 * @brief Convert a trigger-channel to the aquivalent pattern.
			 */
			long triggerSourceToPattern( int channel ) const {
				int nbrExtTrigsPerModule=0, nbrIntTrigsPerModule=0, nbrModules=0;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrModulesInInstrument", &nbrModules ) );
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrInternalTriggers", &nbrIntTrigsPerModule ) );
				nbrIntTrigsPerModule = nbrIntTrigsPerModule / nbrModules;
				errorMessage( AcqrsD1_getInstrumentInfo( _inst, ( char* )"NbrExternalTriggers", &nbrExtTrigsPerModule ) );
				nbrExtTrigsPerModule = nbrExtTrigsPerModule / nbrModules;
				int srcpattern=0;
				long moduleNbr, inputNbr;
				if ( channel > 0 && nbrIntTrigsPerModule > 0 ) {
					//qDebug( "\t Internal" );
					moduleNbr = ( channel - 1 ) / nbrIntTrigsPerModule;
					inputNbr = ( channel - 1 ) % nbrIntTrigsPerModule;
					srcpattern = ( moduleNbr<<16 ) + ( 0x1<<inputNbr );
				} else if ( channel < 0 && nbrExtTrigsPerModule > 0 ) {
					//qDebug( "\t External" );
					channel = -channel;
					moduleNbr = ( channel - 1 ) / nbrExtTrigsPerModule;
					inputNbr = ( channel - 1 ) % nbrExtTrigsPerModule;
					srcpattern = ( moduleNbr<<16 ) + ( 0x80000000>>inputNbr );
				}
				return srcpattern;
			}

			qint32 errorMessage( ViStatus status, QString comment="" ) const { return NSAcqiris::errorMessage( _inst, status, comment ); }
	};


	class ConverterThread : public QThread
	{
		Q_OBJECT

		DigitizerThread* _digitizer;

		bool _stop;

		public:
			ConverterThread( DigitizerThread* digitizer, QObject* p )
				: QThread( p )
				, _digitizer( digitizer )
				, _stop( false )
			{
			}

			void run() {
				prctl(PR_SET_NAME, "ConverterThread");

				qDebug( " ! A ConverterThread is now running with pid %lu!", syscall(SYS_gettid) );

				while ( ! _stop ) {
					DigitizerData* tmp = _digitizer->first();
					if ( tmp == 0 )
						msleep( 100 );
					else {
						for ( int s=0; s<tmp->segments; ++s ) {
							bool pass=true;
							for ( int i=0; i<tmp->channels; ++i ) {
								int rawmax=-INT_MAX, rawmin=INT_MAX;
								DigitizerChannelData* ch = &( tmp->data[ i ] );

								int offset = s * ch->dataDesc->returnedSamplesPerSeg;
								if ( tmp->dataBytes == 1 ) {
									char* d = static_cast<char*>( ch->dataArray );
									//qDebug( "ch->dataArray = %p", ch->dataArray );
									for ( int j=0; j<ch->dataDesc->returnedSamplesPerSeg; ++j ) {
										if (rawmin > d[offset+j])
											rawmin = d[offset+j];
										if (rawmax < d[offset+j])
											rawmax = d[offset+j];
									}
								} else {
									qint16* d = static_cast<qint16*>( ch->dataArray );
									for ( int j=0; j<ch->dataDesc->returnedSamplesPerSeg; ++j ) {
										if (rawmin > d[offset+j])
											rawmin = qFromLittleEndian<qint16>( d[offset+j] );
										if (rawmax < d[offset+j])
											rawmax = qFromLittleEndian<qint16>( d[offset+j] );
									}
								}

								double vGain = ch->dataDesc->vGain;
								double vOffset = ch->dataDesc->vOffset;
								ch->rawmin = rawmin * vGain - vOffset;
								ch->rawmax = rawmax * vGain - vOffset;

								bool pass_min = false;
								bool pass_max = false;
								if ( ch->filter.min && ch->filter.min_min < ch->rawmin && ch->filter.min_max > ch->rawmin )
									pass_min = true;
								if ( ch->filter.max && ch->filter.max_min < ch->rawmax && ch->filter.max_max > ch->rawmax )
									pass_max = true;
								pass &= ( pass_min == ch->filter.min ) && ( pass_max == ch->filter.max );
							}

							if ( pass && tmp->data[0].dataDesc ) {
								int samples = tmp->data[ 0 ].dataDesc->returnedSamplesPerSeg;
								NSCore::AcqisitionDataSet data( tmp->channels, samples );

								data.addResult(  "horizontalposition", static_cast<AqSegmentDescriptor*>( tmp->data[ 0 ].segDescArray )[ s ].horPos );
								for ( int i=0; i<tmp->channels; ++i ) {
									DigitizerChannelData* ch = &( tmp->data[ i ] );
									double vGain = ch->dataDesc->vGain;
									double vOffset = ch->dataDesc->vOffset;

									double* realData = data.getRawChannel( i );
									int offset = s * ch->dataDesc->returnedSamplesPerSeg;
									if ( tmp->dataBytes == 1 ) {
										char* d = static_cast<char*>( ch->dataArray );
										for ( int j=0; j<samples; ++j ) {
											realData[ j ] = d[ offset + j ] * vGain - vOffset;
										}
									} else {
										qint16* d = static_cast<qint16*>( ch->dataArray );
										for ( int j=0; j<samples; ++j ) {
											realData[ j ] = ( qFromLittleEndian<qint16>( d[ offset + j ] ) ) * vGain - vOffset;
										}
									}

									//data.addResult( QString( "channel %1/offset" ).arg( i ), 0 );
									data.addResult( QString( "channel %1/offset" ).arg( i ), vOffset );
									//data.addResult( QString( "channel %1/fullscale" ).arg( i ), 256 * tmp->dataBytes );
									data.addResult( QString( "channel %1/fullscale" ).arg( i ), vGain * pow( 2, 8*tmp->dataBytes ) );
									data.addResult( QString( "channel %1/rawmin" ).arg( i ), ch->rawmin );
									data.addResult( QString( "channel %1/rawmax" ).arg( i ), ch->rawmax );
								}
								data.updateMinMax();
								emit newData( data );
							}
						}
						DigitizerDataPool::the()->deleteData( tmp );
					}
				}
			}

			void stopThread() { _stop = true; }
		signals:
			void newData( NSCore::AcqisitionDataSet );
	};

}; // namespace NSAcqiris

#endif // ACQIRIS_THREAD_H

